
import filecmp
import os
import pathlib

from adbutils import AdbDevice


# todo: make independent of already present stuff on the phone
def test_pull_push_dirs(
        device: AdbDevice,
        device_tmp_dir_path: str,
        local_src_in_dir: pathlib.Path,
        tmp_path: pathlib.Path,
):
    def are_dir_trees_equal(dir1, dir2):
        """
        Compare two directories recursively. Files in each directory are
        assumed to be equal if their names and contents are equal.

        NB: retreived from: https://stackoverflow.com/a/6681395

        @param dir1: First directory path
        @param dir2: Second directory path

        @return: True if the directory trees are the same and 
            there were no errors while accessing the directories or files, 
            False otherwise.
        """

        dirs_cmp = filecmp.dircmp(dir1, dir2)
        if len(dirs_cmp.left_only) > 0 or len(dirs_cmp.right_only) > 0 or \
                len(dirs_cmp.funny_files) > 0:
            return False
        (_, mismatch, errors) = filecmp.cmpfiles(
            dir1, dir2, dirs_cmp.common_files, shallow=False)
        if len(mismatch) > 0 or len(errors) > 0:
            return False
        for common_dir in dirs_cmp.common_dirs:
            new_dir1 = os.path.join(dir1, common_dir)
            new_dir2 = os.path.join(dir2, common_dir)
            if not are_dir_trees_equal(new_dir1, new_dir2):
                return False
        return True

    local_src_out_dir1 = tmp_path / 'dir1'
    local_src_out_dir2 = tmp_path / 'dir2'

    # TODO: push src support dir
    # device.push(local_src_in_dir, device_tmp_dir_path)
    device.adb_output("push", str(local_src_in_dir), device_tmp_dir_path)

    device.sync.pull_dir(device_tmp_dir_path, local_src_out_dir1)

    assert local_src_out_dir1.exists()
    assert local_src_out_dir1.is_dir()

    are_dir_trees_equal(local_src_in_dir, local_src_out_dir1)

    device.sync.pull(device_tmp_dir_path, local_src_out_dir2)

    assert local_src_out_dir2.exists()
    assert local_src_out_dir2.is_dir()

    are_dir_trees_equal(local_src_in_dir, local_src_out_dir2)



import pytest

import adbutils
from adbutils import AdbDevice, Network, BrightnessMode
from adbutils.errors import AdbSyncError


def test_shell(device: AdbDevice):
    for text in ("foo", "?", "&", "but    123"):
        output = device.shell(['echo', '-n', text])
        assert output == text

    output = device.shell("pwd", rstrip=False)
    assert output in ["/\n", "/\r\n"]


def test_shell_without_encoding(device: AdbDevice):
    output = device.shell("echo -n hello", encoding=None)
    assert output == b"hello"

    ret = device.shell2("echo -n hello", encoding=None)
    assert ret.output == b"hello"


def test_shell_stream(device: AdbDevice):
    # c = device.shell(["echo", "-n", "hello world"], stream=True)
    c = device.open_shell("echo -n 'hello world'")
    output = c.read_until_close()
    assert output == "hello world"


def test_adb_shell_raise_timeout(device: AdbDevice):
    with pytest.raises(adbutils.AdbTimeout):
        device.shell("sleep 10", timeout=1.0)


def test_shell2(device: AdbDevice):
    cmd = "echo -n 'hello'; false"
    res = device.shell2(cmd)
    assert res.output == "hello"
    assert res.returncode == 1
    assert res.command == cmd


def test_get_xxx(device: AdbDevice):
    assert device.get_serialno()
    assert device.get_state() == "device"
    # adb connect device devpath is "unknown"
    # assert device.get_devpath().startswith("usb:")


def test_transport(client: adbutils.AdbClient):
    infos = client.list(extended=True)
    transport_id = infos[0].transport_id
    dev = client.device(transport_id=transport_id)
    files = dev.sync.list('/data/local/tmp')
    assert isinstance(files, list)
    
    forward_list = dev.forward_list()
    assert isinstance(forward_list, list)


def test_framebuffer(device: AdbDevice):
    try:
        im = device.framebuffer()
        assert im.size
    except NotImplementedError:
        pass


import abc
from typing import List, Optional, Union

from adbutils._adb import AdbConnection, Network
from adbutils.sync import Sync
from adbutils._proto import ShellReturn

class AbstractShellDevice(abc.ABC):
    @abc.abstractmethod
    def shell(self, cmd: Union[str, List[str]]) -> str:
        pass

    @property
    @abc.abstractmethod
    def sync(self) -> Sync:
        pass


class AbstractDevice(abc.ABC):
    @abc.abstractmethod
    def shell(self, cmd: str, stream: bool = False) -> Union[str, AdbConnection]:
        pass
    
    @abc.abstractmethod
    def shell2(self, cmd: str) -> ShellReturn:
        pass

    @property
    @abc.abstractmethod
    def sync(self) -> Sync:
        pass

    @abc.abstractmethod
    def app_start(self, package_name: str, activity: Optional[str] = None):
        pass

    @abc.abstractmethod
    def uninstall(self, package_name: str):
        pass

    @abc.abstractmethod
    def install_remote(self, path: str, clean: bool = False, flags: list = ["-r", "-t"]):
        pass
    
    @abc.abstractmethod
    def remove(self, path: str):
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Apr 07 2024 20:07:44 by codeskyblue
"""

import logging
import re
import tempfile
import time
import typing
from pathlib import Path
from typing import Optional, Union

import requests

from adbutils._interfaces import AbstractDevice
from adbutils._utils import ReadProgress, humanize
from adbutils.errors import AdbInstallError

logger = logging.getLogger(__name__)


class InstallExtension(AbstractDevice):
    @staticmethod
    def download_apk(url: str, path: Path):
        """
        Download apk file from url
    
        Args:
            url (str): The URL of the APK file to download.
            path (Path): The local file path where the APK will be saved.
    
        Raises:
            requests.exceptions.RequestException: If the download fails.
        """
        try:
            response = requests.get(url, stream=True)
            response.raise_for_status()  # Raise an error for HTTP errors
    
            # Write the content to the specified path
            with open(path, "wb") as file:
                for chunk in response.iter_content(chunk_size=10240):
                    if chunk:  # Filter out keep-alive chunks
                        file.write(chunk)
        except requests.exceptions.RequestException as e:
            logger.warning(f"Failed to download APK from {url}: {e}")
            raise
    

    def install(self,
                path_or_url: Union[str, Path],
                nolaunch: bool = False,
                uninstall: bool = False,
                silent: bool = False,
                callback: typing.Callable[[str], None] = None,
                flags: list = ["-r", "-t"]):
        try:
            import apkutils
            has_apkutils = True
        except ImportError:
            logger.warning("apkutils is not installed, install it with 'pip install adbutils[apk]'")
            has_apkutils = False
        self._install(path_or_url, nolaunch, uninstall, silent, callback, flags, has_apkutils)

    def _install(self,
                path_or_url: Union[str, Path],
                nolaunch: bool = False,
                uninstall: bool = False,
                silent: bool = False,
                callback: typing.Callable[[str], None] = None,
                flags: list = ["-r", "-t"],
                has_apkutils: bool = True):
        """
        Install APK to device

        Args:
            path_or_url: local path or http url
            nolaunch: do not launch app after install
            uninstall: uninstall app before install
            silent: disable log message print
            callback: only two event now: <"BEFORE_INSTALL" | "FINALLY">
            flags (list): default ["-r", "-t"]
            has_apkutils: whether apkutils is installed

        Raises:
            AdbInstallError, BrokenPipeError
        """
        def dprint(msg):
            if not silent:
                print(msg)
                
        if isinstance(path_or_url, str) and re.match(r"^https?://", path_or_url):
            tmpfile = tempfile.NamedTemporaryFile(suffix=".apk")
            self.download_apk(path_or_url, Path(tmpfile.name))
            tmpfile.flush()
            tmpfile.seek(0)
            src_path = Path(tmpfile.name)
            dprint(f"download apk to {src_path}")
        else:
            src_path = Path(path_or_url)
        if not src_path.is_file():
            raise FileNotFoundError(f"File or URL not found: {path_or_url}")
        
        package_name = None
        main_activity = None
        
        if has_apkutils:
            import apkutils
            with apkutils.APK.from_file(str(src_path)) as apk:
                activities = apk.get_main_activities()
                main_activity = activities[0] if activities else None
                package_name = apk.get_package_name()
                if main_activity and main_activity.find(".") == -1:
                    main_activity = "." + main_activity
            
            dprint(f"APK packageName: {package_name}")
            dprint(f"APK mainActivity: {main_activity}")

        device_dst = f"/data/local/tmp/{package_name or 'unknown'}.apk"
        dprint(f'push apk to device: {device_dst}')
        self._push_apk(src_path, device_dst, show_progress=not silent)

        info = self.sync.stat(device_dst)
        apk_size = src_path.stat().st_size
        if not info.size == apk_size:
            AdbInstallError(f'pushed apk size not matched, expect {apk_size} got {info.size}')
        
        if uninstall and package_name:
            dprint(f"uninstall app: {package_name}")
            self.uninstall(package_name)

        dprint("install to android system ...")
        try:
            start = time.time()
            if callback:
                callback("BEFORE_INSTALL")

            self.install_remote(device_dst, clean=True, flags=flags)
            time_used = time.time() - start
            dprint(f"successfully installed, time used {time_used:.1f} seconds")
            if not nolaunch and package_name and main_activity:
                dprint("launch app: %s/%s" % (package_name, main_activity))
                self.app_start(package_name, main_activity)

        except AdbInstallError as e:
            if package_name and e.reason in [
                "INSTALL_FAILED_PERMISSION_MODEL_DOWNGRADE",
                "INSTALL_FAILED_UPDATE_INCOMPATIBLE",
                "INSTALL_FAILED_VERSION_DOWNGRADE"
            ]:
                dprint("uninstall %s because %s" % (package_name, e.reason))
                self.uninstall(package_name)
                self.install_remote(device_dst, clean=True, flags=flags)
                dprint(f"successfully installed, time used {time.time() - start} seconds")
                if not nolaunch and main_activity:
                    dprint(f"Launch app: {package_name}/{main_activity}")
                    self.app_start(package_name, main_activity)
            else:
                # print to console
                print(
                    "Failure " + e.reason + "\n" +
                    "Remote apk is not removed. Manually install command:\n\t"
                    + "adb shell pm install -r -t " + device_dst)
                raise
        finally:
            if callback:
                callback("FINALLY")

    def _push_apk(self, apk_path: Path, device_dst: str, show_progress: bool = True):
        """
        Push APK file to device with progress indication.
    
        Args:
            apk_path (Path): Path to the APK file.
            device_dst (str): Destination path on the device.
    
        Returns:
            None
        """
        start = time.time()
        length = apk_path.stat().st_size
        with apk_path.open("rb") as fd:
            if show_progress:
                r = ReadProgress(fd, length, source_path=str(apk_path))
                self.sync.push(r, device_dst)
            else:
                self.sync.push(fd, device_dst)
        logger.info("Success pushed, time used %d seconds" % (time.time() - start))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Created on Fri May 06 2022 10:33:39 by codeskyblue
"""

from __future__ import annotations

from typing import Optional

from adbutils._adb import BaseClient
from adbutils._deprecated import DeprecatedExtension
from adbutils._device_base import BaseDevice
from adbutils._proto import *
from adbutils._version import __version__
from adbutils.install import InstallExtension
from adbutils.screenrecord import ScreenrecordExtension
from adbutils.screenshot import ScreenshotExtesion
from adbutils.shell import ShellExtension
from adbutils.sync import Sync


class AdbDevice(
    BaseDevice,
    ShellExtension,
    ScreenrecordExtension,
    ScreenshotExtesion,
    InstallExtension,
    DeprecatedExtension,
):
    """provide custom functions for some complex operations"""

    def __init__(
        self, client: BaseClient, serial: Optional[str] = None, transport_id: Optional[int] = None
    ):
        BaseDevice.__init__(self, client, serial, transport_id)
        ScreenrecordExtension.__init__(self)
    
    @property
    def sync(self) -> Sync:
        return Sync(self)



"""
extra functions test

效果生效与否不好判定（例如屏幕亮暗），部分用例仅作冒烟测试
"""

import os
import io
import pathlib
import re
import time

import uuid

import pytest

from adbutils import AdbDevice, Network, BrightnessMode
from adbutils.errors import AdbSyncError


def test_battery(device: AdbDevice):
    print(device.battery().level)


def test_keyevent(device: AdbDevice):
    # make sure no error raised
    device.keyevent(4)
    device.volume_up(2)
    device.volume_down(3)
    device.volume_mute()


def test_brightness(device: AdbDevice):
    current_brightness = device.brightness_value
    device.brightness_value = 100
    assert device.brightness_value == 100
    device.brightness_value = current_brightness

    current_mode = device.brightness_mode
    if current_mode == BrightnessMode.AUTO:
        device.brightness_mode = BrightnessMode.MANUAL
        assert device.brightness_mode == BrightnessMode.MANUAL
    elif current_mode == BrightnessMode.MANUAL:
        device.brightness_mode = BrightnessMode.AUTO
        assert device.brightness_mode == BrightnessMode.AUTO
    device.brightness_mode = current_mode


def test_switch_screen(device: AdbDevice):
    device.switch_screen(False)
    device.switch_screen(True)


def test_switch_airplane(device: AdbDevice):
    device.switch_airplane(True)
    device.switch_airplane(False)


def test_switch_wifi(device: AdbDevice):
    device.switch_wifi(False)
    device.switch_wifi(True)


def test_swipe(device: AdbDevice):
    device.swipe(100, 100, 400, 400)


def test_click(device: AdbDevice):
    device.click(100, 100)


def test_send_keys(device: AdbDevice):
    device.send_keys("1234")


def test_wlan_ip(device: AdbDevice):
    device.switch_airplane(False)
    device.switch_wifi(True)
    time.sleep(3)
    ip = device.wlan_ip()
    assert ip, 'ip is empty'


def test_app_start_stop(device: AdbDevice):
    d = device
    package_name = "io.appium.android.apis"
    if package_name not in d.list_packages():
        pytest.skip(package_name + " should be installed, to start test")
    d.app_start(package_name)
    time.sleep(1)
    assert device.app_current().package == package_name
    d.app_stop(package_name)
    time.sleep(.5)
    assert device.app_current().package != package_name


def test_sync_pull_push(device: AdbDevice, device_tmp_path, tmp_path: pathlib.Path):
    src = io.BytesIO(b"Hello 1")
    device.sync.push(src, device_tmp_path)
    assert b"Hello 1" == device.sync.read_bytes(device_tmp_path)

    device.sync.push(b"Hello 12", device_tmp_path)
    assert "Hello 12" == device.sync.read_text(device_tmp_path)

    target_path = tmp_path / "hi.txt"
    target_path.write_text("Hello Android")
    dst_path = tmp_path / "dst.txt"
    dst_path.unlink(missing_ok=True)

    device.sync.push(target_path, device_tmp_path)
    assert "Hello Android" == device.sync.read_text(device_tmp_path)
    device.sync.pull(device_tmp_path, dst_path)
    assert "Hello Android" == dst_path.read_text(encoding="utf-8")

    data = b""
    for chunk in device.sync.iter_content(device_tmp_path):
        data += chunk
    assert b"Hello Android" == data


def test_sync_pull_file_push(device: AdbDevice, device_tmp_path, tmp_path: pathlib.Path):
    src = io.BytesIO(b"Hello 1")
    device.sync.push(src, device_tmp_path)
    assert b"Hello 1" == device.sync.read_bytes(device_tmp_path)

    device.sync.push(b"Hello 12", device_tmp_path)
    assert "Hello 12" == device.sync.read_text(device_tmp_path)

    target_path = tmp_path / "hi.txt"
    target_path.write_text("Hello Android")
    dst_path = tmp_path / "dst.txt"
    dst_path.unlink(missing_ok=True)

    device.sync.push(target_path, device_tmp_path)
    assert "Hello Android" == device.sync.read_text(device_tmp_path)
    device.sync.pull_file(device_tmp_path, dst_path)
    assert "Hello Android" == dst_path.read_text(encoding="utf-8")

    data = b""
    for chunk in device.sync.iter_content(device_tmp_path):
        data += chunk
    assert b"Hello Android" == data


def test_sync_push_to_dir(device: AdbDevice, device_tmp_dir, tmp_path: pathlib.Path):
    random_data = str(uuid.uuid4()).encode()
    src = io.BytesIO(random_data)
    with pytest.raises(AdbSyncError):
        device.sync.push(src, device_tmp_dir)
    src_path = tmp_path.joinpath("random.txt")
    src_path.write_bytes(random_data)
    assert device.sync.push(src_path, device_tmp_dir) == len(random_data)
    assert random_data == device.sync.read_bytes(device_tmp_dir + "/random.txt")


def test_screenshot(device: AdbDevice):
    im = device.screenshot()
    assert im.mode == "RGB"


def test_app_info(device: AdbDevice):
    pinfo = device.app_current()
    app_info = device.app_info(pinfo.package)
    assert app_info.package_name is not None


def test_window_size(device: AdbDevice):
    w, h = device.window_size()
    assert isinstance(w, int)
    assert isinstance(h, int)
    
    is_landscape = device.rotation() % 2 == 1
    nw, nh = device.window_size(not is_landscape)
    assert w == nh and h == nw


def test_is_screen_on(device: AdbDevice):
    bool_result = device.is_screen_on()
    assert isinstance(bool_result, bool)


def test_open_browser(device: AdbDevice):
    device.open_browser("https://example.org")


def test_dump_hierarchy(device: AdbDevice):
    output = device.dump_hierarchy()
    assert output.startswith("<?xml")
    assert output.rstrip().endswith("</hierarchy>")


def test_remove(device: AdbDevice):
    remove_path = "/data/local/tmp/touch.txt"
    device.shell(["touch", remove_path])
    assert device.sync.exists(remove_path)
    device.remove(remove_path)
    assert not device.sync.exists(remove_path)


# def test_create_connection(device: AdbDevice, device_tmp_path: str):
#     device.sync.push(b"hello", device_tmp_path)
#     device.create_connection(Network.LOCAL_FILESYSTEM, device_tmp_path)

def test_logcat(device: AdbDevice, tmp_path: pathlib.Path):
    logcat_path = tmp_path / "logcat.txt"
    logcat = device.logcat(logcat_path, clear=True, command="logcat -v time", re_filter="I/TAG")
    device.shell(["log", "-p", "i", "-t", "TAG", "hello"])
    time.sleep(.1)
    logcat.stop()
    assert logcat_path.exists()
    assert re.compile(r"I/TAG.*hello").search(logcat_path.read_text(encoding="utf-8"))

